#include <stdio.h>
#include <stdlib.h>
#include <opencv2/opencv.hpp>     

using namespace cv;

const char* kNamedWindow = "OpenCV Challenge 3 Solution";
const int kWAIT_LEFT = 65361;
const int kWAIT_RIGHT = 65363;
const int kWAIT_SPACE = 32;

void debug_show(const Mat& out,const char text[])
{
  //imshow(kNamedWindow, out);
  //printf("%s\n",text);
  //int key = waitKey(0);
}

void debug_show(const Mat& data,const Mat& out,const Rect& outRect,const char text[])
{
  data.copyTo(Mat(out,outRect));
  debug_show(out,text);
}

void debug_show_gray(const Mat& data,const Mat& out,const Rect& outRect,const char text[])
{
  static Mat temp;
  cvtColor( data, temp, CV_GRAY2BGR );
  debug_show(temp,out,outRect,text);
}

void jjHoughCircle(const Mat& data, Mat& out, double min_r, double max_r, string& message)
{
  int size = data.rows; if( data.cols < size ) size = data.cols;
  int min = (int)(min_r*size/2);
  int max = (int)(max_r*size/2);

  char msg_buff[512];
  sprintf(msg_buff,"  Hough Circles (%d,%d)",min,max);
  message += string(msg_buff);

  vector<Vec3f> circle_data;
  HoughCircles(data, circle_data, CV_HOUGH_GRADIENT, 1, 300,
               20, 30, min, max);

  for( size_t i = 0; i < circle_data.size(); i++ )
  {
    Vec3i c = circle_data[i];
    circle( out, Point(c[0], c[1]), c[2], Scalar(0,0,255), 1, CV_AA);
    circle( out, Point(c[0], c[1]), 2, Scalar(0,255,0), 3, CV_AA);
    sprintf(msg_buff,"  Circle %d has radius: %d\n",i,c[2]);
    message += string(msg_buff);
  }
}

void jjContourRedraw(const Mat& data, Mat& out, int line_size)
{
  vector<vector<Point> > lines;
  vector<Vec4i> hierarchy;
  findContours( data, lines, hierarchy, CV_RETR_EXTERNAL, CV_CHAIN_APPROX_SIMPLE, Point(0,0) );

  cvtColor( data, out, CV_GRAY2BGR);
  RNG rng(12346); // Random Number Generator - For Rainbow Colors;
  for( int i = 0; i< lines.size(); i++ )
  {
    Scalar color = Scalar( rng.uniform(0, 255), rng.uniform(0,255), rng.uniform(0,255) );
    drawContours( out, lines, i, color, line_size, CV_AA, hierarchy, 0, Point() );
  }
}

/**
 * @param inData - input
 * @param out - output from the pipeline after step #
 * @param step - how far in the pipeline to run.
 */
void run_pipeline(const Mat& data, Mat& out, int step, string &message)
{
  cvtColor( data, out, CV_BGR2GRAY, 1 );
  message = "cvtColor - CV_BGR2GRAY, 1";
  if( step < 2 ) return;

  blur( out, out, Size(3,3) );
  message = "blur - Size(3,3)";
  if( step < 3 ) return;
  Mat blur; out.copyTo(blur); // Save for hough Circles.

  Canny( out, out, 20, 60 );
  message = "canny - 20 / 60";
  if( step < 4 ) return;

  jjContourRedraw( out, out, 2 );
  message = string("findContours - CV_RETR_EXTERNAL, CV_CHAIN_APPROX_SIMPLE\n")
          + string("  drawContours - 2px wide.");
  if( step < 5 ) return;

  cvtColor( out, out, CV_BGR2GRAY, 1 );
  jjContourRedraw( out, out, 1 );
  message = string("contours, 2nd step\n")
          + string("  findContours - CV_RETR_EXTERNAL, CV_CHAIN_APPROX_SIMPLE\n")
          + string("  drawContours - 1px wide.");
  if( step < 6 ) return;

  message = string("Hough Circles (r:80-100%) from Blur\n");
  jjHoughCircle( blur, out, .80, 1, message );
  if( step < 7 ) return;
  
}

void seperate_symbols(const Mat& data, const Mat& out, const Rect& outRect, int num_steps, string& mess)
{
  Mat temp;
  run_pipeline(data,temp,num_steps,mess);
  if( temp.channels() == 1 )
    cvtColor( temp, temp, CV_GRAY2BGR);
  
  temp.copyTo(Mat(out,outRect));
}

void analyzeImage(const Mat& img, int num_steps, bool debug_msg, bool tall)
{
  namedWindow(kNamedWindow, CV_WINDOW_AUTOSIZE );

  int rows = img.rows;
  int cols = img.cols;
  Rect copy_size, left_size, right_size;
  Rect out_size, out_left, out_right, out_left2, out_right2;

  if( tall ) {
    copy_size = Rect(0,0,450*cols/rows,450);
    left_size = Rect(0,0,copy_size.width/2,copy_size.height);
    right_size = Rect(copy_size.width/2,0,copy_size.width/2,copy_size.height);

    out_size =  Rect(0,0, left_size.width*4+30, left_size.height);
    out_left = left_size;
    out_right = Rect(right_size.x+10,right_size.y,right_size.width,right_size.height);
    out_left2 =
      Rect(left_size.width*2+20, left_size.y, out_left.width, out_left.height);
    out_right2 =
      Rect(left_size.width*3+30, left_size.y, out_left.width, out_left.height);
  }
  else {
    copy_size = Rect(0,0,300*cols/rows,300);
    left_size = Rect(0,0,copy_size.width/2,copy_size.height);
    right_size = Rect(copy_size.width/2,0,copy_size.width/2,copy_size.height);

    out_size =  Rect(0,0, left_size.width+10+right_size.width, left_size.height*2+10);
    out_left = left_size;
    out_right = Rect(right_size.x+10,right_size.y,right_size.width,right_size.height);
    out_left2 =
      Rect(out_left.x, out_left.y+10+out_left.height, out_left.width, out_left.height);
    out_right2 =
      Rect(out_right.x, out_right.y+10+out_right.height, out_right.width, out_right.height);

  }
   
  Scalar fill = Scalar(128,40,200);

  Mat copy;
  resize(img,copy,Size(copy_size.width,copy_size.height));
  Mat left=Mat(copy,left_size);
  Mat right=Mat(copy,right_size);

  Mat out = Mat(out_size.height,out_size.width,copy.type(),fill);
  left.copyTo(Mat(out,out_left));
  right.copyTo(Mat(out,out_right));

  string message;
  seperate_symbols(left,out,out_left2,num_steps,message);
  if( debug_msg ) printf("Left: %s\n",message.c_str());
  seperate_symbols(right,out,out_right2,num_steps,message);
  if( debug_msg ) printf("Right: %s\n",message.c_str());

  imshow(kNamedWindow, out);
}

const int MAX_LEVEL = 9;
const int MIN_LEVEL = 1;

int main(int argc, char** argv )
{
  printf("usage: Solution [<Image_Path>]\n");

  if( argc == 2 )
  {
    Mat image = imread( argv[1], 1 );

    if ( !image.data )
    {
      printf("No image data in file '%s'.\n",argv[1]);
      return -1;
    }

    int level = MIN_LEVEL;
    for(;;)
    {
      analyzeImage(image,level,true,false);
      char c = (char)waitKey(0);
      if( c == 27 ) break;
      switch(c) {
      case 'n': level++; if(level>MAX_LEVEL) level=MAX_LEVEL; break;
      case 'b': level--; if(level<MIN_LEVEL) level=MIN_LEVEL; break;
      }
    }
  }

  else
  {
    VideoCapture cap;
    int camNum = 0;
    cap.open(camNum);
    if( !cap.isOpened() )
    {
      printf("Could not open camera %d\n",camNum);
      return -1;
    }
    Mat frame, image;

    int level = MIN_LEVEL;
    bool show_msg = true;
    for(;;)
    {
      cap >> frame;
      analyzeImage(frame,level,show_msg,true);
      show_msg = false;
      char c = (char)waitKey(10);
      if( c == 27 ) break;
      switch(c) {
      case 'n': level++; if(level>MAX_LEVEL) level=MAX_LEVEL; show_msg = true; break;
      case 'b': level--; if(level<MIN_LEVEL) level=MIN_LEVEL; show_msg = true; break;
      }
    }
  }
   
  return 0;
}

